#ifndef REFLEX_CAMERA_H
#define REFLEX_CAMERA_H

#include <reflex/base/FileAccessObject.hpp>
#include <reflex/driver/ReflexDriverDev.hpp>
#include <reflex/camera/CameraIntrinsics.hpp>
#include <reflex/camera/CameraExtrinsics.hpp>
#include <reflex/camera/ImageRectifier.hpp>
#include <reflex/camera/SensorConfig.hpp>
#include <reflex/system/UnitInfo.hpp>
#include <reflex/system/UnitGeometry.hpp>
#include <reflex/system/Environment.hpp>


namespace reflex
{

class ReflexCamera : public FileAccessObject
{
public:

    class CameraSensor
    {
    public:

        CameraSensor();
        virtual ~CameraSensor();

        CameraIntrinsics          intrinsics;
        RectificationParameters   rectification;
        ImageRectifier            rectifier;
        SensorConfig              sensorConfig;
    };


    ReflexCamera();
    ReflexCamera(const string& ipAddress, const string& serialNumber = string(""));
    virtual ~ReflexCamera();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    int getNbSystems() const { return this->unitInfo.systems.size(); }
    int getNbSensors() const { return this->unitInfo.sensors.size(); }

    int init(const string& dataFolder = "/tmp/");
    int loadCalibrationData(const string& dataFolder);
    int loadRectifier();

    int reboot();

    int changeOperationState(int state, float timeout = 5.);
    int changeNetworkSettings(const std::string& ipAddress, const std::string& netmask = "255.255.255.0", const std::string& gateway = "");

    int changeSensorMode(int sensorID, int8_t mode);
    int changeSensorShutter(int sensorID, uint32_t shutter);
    int changeSensorGain(int sensorID, float gain);

    int receiveData(ReflexRawData& data) const;
    int receiveData(ReflexData& data) const;

    int receiveImage(int sensorID, ReflexImageCV& image) const;
    int receiveImages(const std::vector<int>& sensorID, std::vector<ReflexImageCV>& images) const;
    int receiveImagesFull(std::vector<ReflexImageCV>& images) const;

    int receiveDataAndImage(ReflexData& data, int sensorID, ReflexImageCV& image) const;
    int receiveDataAndImages(ReflexData& data, const std::vector<int>& sensorID,  std::vector<ReflexImageCV>& image) const;

    int receiveSystemStatus(int systemID, ReflexSystemStatus& status) const;
    int receiveSystemStatus(std::vector<ReflexSystemStatus>& status) const;

    int receiveSensorStatus(int sensorID, ReflexSensorStatus& status) const;
    int receiveSensorStatus(std::vector<ReflexSensorStatus>& status) const;

    string getDataFolder() const { return _dataFolder; }

    static int getNIRSystemID(int unitType);
    static int getRGBSystemID(int unitType);

    int                               unitType;

    int                               operationState;

    string                            ipAddress;
    string                            serialNumber;

    string                            softwareVersion;
    int                               softwareVersionMajor;
    int                               softwareVersionMinor;
    int                               softwareVersionPatch;

    UnitInfo                          unitInfo;
    UnitGeometry                      unitGeometry;
    Environment                       environment;
    vector<vector<CameraExtrinsics>>  extrinsics;

    vector<CameraSensor>              cameras;

protected:

    bool _initialized;
    string _dataFolder;

};

} // end namespace reflex


#endif
