#ifndef ACCUMULATOR_ARRAY_H
#define ACCUMULATOR_ARRAY_H

#include <reflex/base/FileAccessObject.hpp>


namespace reflex
{

template<class Tp>
class AccumulatorArray : public FileAccessObject
{
public:

    enum types
    {
        ACCUMULATOR_ARRAY_UNDEFINED = -1,
        ACCUMULATOR_ARRAY_1D_LIN,
        ACCUMULATOR_ARRAY_1D_LOG,
        ACCUMULATOR_ARRAY_2D_LIN,
        ACCUMULATOR_ARRAY_2D_LOG
    };

    AccumulatorArray(int type);
    virtual ~AccumulatorArray();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);
};


/****************************************************************************************************/

template<class Tp>
class AccumulatorArray1 : public AccumulatorArray<Tp>
{
public:

    AccumulatorArray1(const AccumulatorArray1& origin);
    AccumulatorArray1(int type);
    AccumulatorArray1(int type, Tp vMin, Tp vMax, int nbBins);
    virtual ~AccumulatorArray1();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    bool empty() const { return nbBins == 0; }

    virtual void initialize();
    virtual void initialize(Tp vMin, Tp vMax, int nbBins);
    virtual void initialize(const AccumulatorArray1<Tp>& acc, bool copyContent = false);

    virtual Tp getBin(double value) const = 0;
    virtual Tp getValue(double bin) const = 0;

    virtual double getWeightFromBin(int bin) const;
    //virtual double getWeightFromValue(double val) const;
    //virtual bool setWeight(int bin, double weight);
    //virtual void add(double value) { histogram += value; }
    //virtual void add(double value, double weight, int interpolation = CV_INTER_NN);
    //virtual bool add(const AccumulatorArray1& acc);
    //virtual void scale(double factor) { histogram *= factor; }
    //void getMaximum(int* maxBin, double* maxValue, double* maxWeight, int filterSize = 1) const;
    //void getMaxima(vector<int>& maxBin, vector<double>& maxValue, vector<double>& maxWeight, int n, int filterSize = 1) const;
    //int getMaximumBin(int filterSize = 1) const;
    //double getMaximumValue(int filterSize = 1) const;
    //double getMaximumWeight(int filterSize = 1) const;

    void blurAverage(int filterSize, int borderType = cv::BORDER_DEFAULT);
    void blurGaussian(int filterSize, int borderType = cv::BORDER_DEFAULT);
    void blurMedian(int filterSize);

    void normalize();
    bool normalize(const AccumulatorArray1& histogram);

    //virtual cv::Mat draw(int width, int border = 16, bool markMaximum = false) const;
    //virtual cv::Mat draw(const cv::Size& size, int border = 16, bool markMaximum = false) const;

    cv::Mat histogram;
    Tp valueMin;
    Tp valueMax;
    int nbBins;
};


template<class Tp>
class AccumulatorArrayLin1 : public AccumulatorArray1<Tp>
{
public:

    AccumulatorArrayLin1();
    AccumulatorArrayLin1(Tp vMin, Tp vMax, int nbBins);
    virtual ~AccumulatorArrayLin1();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual Tp getBin(Tp value) const;
    virtual Tp getValue(Tp bin) const;

};

typedef AccumulatorArrayLin1<float>  AccumulatorArrayLin1f;
typedef AccumulatorArrayLin1<double> AccumulatorArrayLin1d;


template<class Tp>
class AccumulatorArrayLog1 : public AccumulatorArray1<Tp>
{
public:

    AccumulatorArrayLog1();
    AccumulatorArrayLog1(Tp vMin, Tp vMax, int nbBins);
    virtual ~AccumulatorArrayLog1();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual void initialize();
    virtual void initialize(Tp vMin, Tp vMax, int nbBins);

    virtual Tp getBin(Tp value) const;
    virtual Tp getValue(Tp bin) const;


    double factor;
};

typedef AccumulatorArrayLog1<float>  AccumulatorArrayLog1f;
typedef AccumulatorArrayLog1<double> AccumulatorArrayLog1d;


/****************************************************************************************************/


template<class Tp>
class AccumulatorArray2 : public AccumulatorArray<Tp>
{
public:

    AccumulatorArray2(const AccumulatorArray2& origin);
    AccumulatorArray2(int type);
    AccumulatorArray2(int type, cv::Point_<Tp> vMin, cv::Point_<Tp> vMax, cv::Size2i nbBins);
    virtual ~AccumulatorArray2();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual void initialize();
    virtual void initialize(cv::Point_<Tp> vMin, cv::Point_<Tp> vMax, cv::Size2i nbBins);
    virtual void initialize(const AccumulatorArray2<Tp>& acc, bool copyContent = false);

    bool empty() const { return nbBins == cv::Size2i(0,0); }

    virtual cv::Point_<Tp> getBin(const cv::Point_<Tp>& value) const = 0;
    virtual cv::Point_<Tp> getValue(const cv::Point_<Tp>& bin) const = 0;

    virtual double getWeightFromBin(const cv::Point2i& bin) const;
    virtual double getWeightFromValue(const cv::Point_<Tp>& val) const;
    virtual bool setWeight(const cv::Point2i& bin, double weight);
    virtual void add(Tp value) { histogram += value; }
    virtual void add(const cv::Point_<Tp>& value, double weight, int interpolation = CV_INTER_NN);
    virtual int  add(const AccumulatorArray2& acc);
    virtual int  add(const cv::Mat& imgWeights, double scale = 1., const cv::Mat& mask = cv::Mat());
    virtual int  add(const cv::Mat& imgWeights, const cv::Rect& roi, double scale = 1., const cv::Mat& mask = cv::Mat());
    virtual int  add(const cv::Mat& imgWeights, const cv::Point2i& origin, double scale = 1., const cv::Mat& mask = cv::Mat());
    virtual void scale(double factor) { histogram *= factor; }

    void normalize();
    bool normalize(const AccumulatorArray2& histogram);

    void blurAverage(const cv::Size2i& filterSize, int borderType = cv::BORDER_DEFAULT);
    void blurGaussian(const cv::Size2i& filterSize, int borderType = cv::BORDER_DEFAULT);
    void blurMedian(int size);

    void getMaximum(cv::Point2i& maxBin, cv::Point_<Tp>& maxValue, double& maxWeight, cv::Size2i filterSize = cv::Size2i(1,1)) const;
    cv::Point2i getMaximumBin(cv::Size2i filterSize = cv::Size2i(1,1)) const;
    cv::Point_<Tp> getMaximumValue(cv::Size2i filterSize = cv::Size2i(1,1)) const;
    double getMaximumWeight(cv::Size2i filterSize = cv::Size2i(1,1)) const;

    virtual cv::Mat draw(double scale = 1., bool markMaximum = false) const;

    cv::Mat histogram;
    cv::Point_<Tp> valueMin;
    cv::Point_<Tp> valueMax;
    cv::Size2i nbBins;
};


template<class Tp>
class AccumulatorArrayLin2 : public AccumulatorArray2<Tp>
{
public:

    AccumulatorArrayLin2();
    AccumulatorArrayLin2(cv::Point_<Tp> vMin, cv::Point_<Tp> vMax, cv::Size2i nbBins);
    virtual ~AccumulatorArrayLin2();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual cv::Point_<Tp> getBin(const cv::Point_<Tp>& value) const;
    virtual cv::Point_<Tp> getValue(const cv::Point_<Tp>& bin) const;

};

typedef AccumulatorArrayLin2<float>  AccumulatorArrayLin2f;
typedef AccumulatorArrayLin2<double> AccumulatorArrayLin2d;


} // end of namespace reflex


#endif
