#ifndef IMAGE_RECTIFIER_H
#define IMAGE_RECTIFIER_H

#include <reflex/base/FileAccessObject.hpp>
#include <reflex/camera/RectificationParameters.hpp>
#include <reflex/camera/CameraIntrinsics.hpp>

namespace reflex
{

class ImageRectifier : public FileAccessObject
{
public:

    ImageRectifier();
    ImageRectifier(const string& fnIntrinsics, const string& fnRectification);

    virtual ~ImageRectifier();

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    int loadParams(const string& fnIntrinsics, const string& fnRectification);

    int rectifyImage(const cv::Mat imgRaw, cv::Mat& imgRectified, int interpolation = cv::INTER_LINEAR, int borderMode = cv::BORDER_CONSTANT, const cv::Scalar& borderValue = cv::Scalar(127)) const;

    int rectifyRows(const cv::Mat imgRaw, cv::Mat& imgRectified, const cv::Mat& imgMask, int interpolation = cv::INTER_LINEAR, int borderMode = cv::BORDER_CONSTANT, const cv::Scalar& borderValue = cv::Scalar(127)) const;

    int rectifyRois(const cv::Mat imgRaw, cv::Mat& imgRectified, const std::vector<cv::Rect>& rois, int interpolation = cv::INTER_LINEAR, int borderMode = cv::BORDER_CONSTANT, const cv::Scalar& borderValue = cv::Scalar(127)) const;
    int rectifyRoi(const cv::Mat imgRaw, cv::Mat& imgRectified, const cv::Rect& roi, int interpolation = cv::INTER_LINEAR, int borderMode = cv::BORDER_CONSTANT, const cv::Scalar& borderValue = cv::Scalar(127)) const;

    template<class Tp>
    int rectifyPoints(const std::vector<cv::Point_<Tp>>& pointsRaw, std::vector<cv::Point_<Tp>>& pointsRectified) const;

    template<class Tp>
    void projectTo3D(const cv::Point_<Tp>& pt2D, cv::Point3_<Tp>& pt3D) const;

    template<class Tp>
    void projectTo3D(const cv::Vec<Tp,2>& pt2D, cv::Vec<Tp,3>& pt3D) const;

    template<class Tp>
    void projectTo3D(const cv::Point_<Tp>& pt2D, Tp disparity, cv::Point3_<Tp>& pt3D) const;

    template<class Tp>
    void projectTo3D(const cv::Vec<Tp,2>& pt2D, Tp disparity, cv::Vec<Tp,3>& pt3D) const;

    CameraIntrinsics intrinsics;
    RectificationParameters params;

};


template<class Tp>
int ImageRectifier::rectifyPoints(const std::vector<cv::Point_<Tp>>& pointsRaw, std::vector<cv::Point_<Tp>>& pointsRectified) const
{
    if(params.map.empty())
        return ERROR_INITIALIZATION;

    if(typeid(Tp) == typeid(int))
    {
        std::vector<cv::Point2f> pointsRaw2f(pointsRaw.begin(), pointsRaw.end());
        std::vector<cv::Point2f> pointsRectified2f;
        cv::undistortPoints(pointsRaw2f, pointsRectified2f, intrinsics.cameraMatrix, intrinsics.distCoeffs, params.R, params.P);
        pointsRectified = std::vector<cv::Point_<Tp>>(pointsRectified2f.begin(), pointsRectified2f.end());
    }
    else
        cv::undistortPoints(pointsRaw, pointsRectified, intrinsics.cameraMatrix, intrinsics.distCoeffs, params.R, params.P);

    return ERROR_NONE;
}


template<class Tp>
void ImageRectifier::projectTo3D(const cv::Point_<Tp>& pt2D, cv::Point3_<Tp>& pt3D) const
{
    pt3D.x = (pt2D.x + params.q03) / params.q23;
    pt3D.y = (pt2D.y + params.q13) / params.q23;
    pt3D.z = 1.;
}


template<class Tp>
void ImageRectifier::projectTo3D(const cv::Vec<Tp,2>& pt2D, cv::Vec<Tp,3>& pt3D) const
{
    pt3D[X] = (pt2D.x + params.q03) / params.q23;
    pt3D[Y] = (pt2D.y + params.q13) / params.q23;
    pt3D[Z] = 1.;
}


template<class Tp>
void ImageRectifier::projectTo3D(const cv::Point_<Tp>& pt2D, Tp disparity, cv::Point3_<Tp>& pt3D) const
{
    const Tp pw = (Tp)(1.0f / (disparity * params.q32 + params.q33));
    pt3D.x = (pt2D.x + params.q03) * pw;
    pt3D.y = (pt2D.y + params.q13) * pw;
    pt3D.z = params.q23 * pw;
}


template<class Tp>
void ImageRectifier::projectTo3D(const cv::Vec<Tp,2>& pt2D, Tp disparity, cv::Vec<Tp,3>& pt3D) const
{
    const Tp pw = (Tp)(1.0f / (disparity * params.q32 + params.q33));
    pt3D[X] = (pt2D.x + params.q03) * pw;
    pt3D[Y] = (pt2D.y + params.q13) * pw;
    pt3D[Z] = params.q23 * pw;
}



} // end namespace reflex

#endif
