#ifndef _REFLEX_DRIVER_H_
#define _REFLEX_DRIVER_H_

#include "reflex.hpp"


namespace reflex
{

struct ReflexCameraUnit
{
    std::string ip_address;
    std::string serial_number;
    int8_t      type;
    int8_t      nb_systems;
    int8_t      nb_sensors;
    int8_t      sw_version_major;
    int8_t      sw_version_minor;
    int8_t      sw_version_patch;

    ReflexCameraUnit()
    {
        type             = -1;
        nb_systems       = -1;
        nb_sensors       = -1;
        sw_version_major = -1;
        sw_version_minor = -1;
        sw_version_patch = -1;
    }
};


struct ReflexSystemStatus
{
    double      timestamp;
    int8_t      system_id;
    float       temperature;
    int16_t     error;
    int8_t      state;

    ReflexSystemStatus()
    {
        timestamp   = -1;
        system_id   = -1;
        temperature = -1;
        error       = ERROR_NONE;
        state       = SYSTEM_STATE_UNDEFINED;
    }
};


struct ReflexSensorStatus
{
    double      timestamp;
    int8_t      sensor_id;
    int16_t     error;
    uint32_t    frame_id;
    float       framerate;
    int8_t      exposure_mode;
    uint32_t    shutter;
    float       gain;

    ReflexSensorStatus()
    {
        timestamp      = -1;
        sensor_id      = -1;
        error          =  ERROR_NONE;
        frame_id       =  0;
        framerate      = -1;
        exposure_mode  = -1;
        shutter        =  0;
        gain           = -1.;
    }
};


struct ReflexContour
{
    std::vector<int16_t> x;
    std::vector<int16_t> y;
};


struct ReflexReflector
{
    ReflexContour contour;
    float         pos2d[2];
    float         pos3d[3];
    float         area;
    float         circularity;
    float         score;

    ReflexReflector()
    {
        pos2d[0] = 0.;
        pos2d[1] = 0.;
        pos3d[0] = 0.;
        pos3d[1] = 0.;
        pos3d[2] = 0.;
        area = -1.;
        circularity = -1.;
        score = -1.;
    }
};


struct ReflexObject
{
    float    pos2d[2];
    float    pos3d[3];
    float    vel3d[3];
    float    score;
    int16_t  id;
    uint32_t lifetime;

    ReflexObject()
    {
        pos2d[0] = 0.;
        pos2d[1] = 0.;
        pos3d[0] = 0.;
        pos3d[1] = 0.;
        pos3d[2] = 0.;
        vel3d[0] = 0.;
        vel3d[1] = 0.;
        vel3d[2] = 0.;
        score = -1.;
        id = -1;
        lifetime = 0;
    }
};


struct ReflexData
{
    double                       timestamp;
    float                        framerate;
    int16_t                      error;
    uint32_t                     frame_id;
    std::vector<ReflexReflector> reflectors;
    std::vector<ReflexObject>    objects;

    ReflexData()
    {
        timestamp = -1.;
        framerate = -1.;
        error     = -1;
        frame_id  = 0;
    }
};


struct ReflexImage
{
    double    timestamp;
    uint32_t  frame_id;
    uint32_t  shutter;
    float     gain;
    int8_t    exposure_mode;

    uint16_t  width;
    uint16_t  height;
    int8_t    type;
    int8_t    sensor_id;

    uint8_t*  channel1;
    uint8_t*  channel2;
    uint8_t*  channel3;

    uint32_t   step;

    ReflexImage()
    {
        timestamp      = -1.;
        frame_id       =  0;
        shutter        =  0;
        gain           = -1.;
        exposure_mode  = -1;
        width          = -1;
        height         = -1;
        type           = -1;
        sensor_id      = -1;
        channel1       = NULL;
        channel2       = NULL;
        channel3       = NULL;
        step           = -1;
    }
};


int reflex_receive_camera_units(std::vector<ReflexCameraUnit>& camera_units);
int reflex_receive_camera_unit_info(const std::string& ip_address, ReflexCameraUnit& camera_unit);
int reflex_receive_system_config(const std::string& ip_address, const std::string& target_directory = std::string("/tmp/"));
int reflex_receive_calibration_data(const std::string& ip_address, const std::string& target_directory = std::string("/tmp/"));

int reflex_reboot_camera_unit(const std::string& ip_address);

int reflex_receive_camera_pose(const std::string& ip_address, std::vector<double>& pose);
int reflex_change_camera_pose(const std::string& ip_address, const std::vector<double>& pose);

int reflex_change_operation_state(const std::string& ip_address, int state);
int reflex_change_network_settings(const std::string& ip_address,
                                   const std::string& new_ip_address,
                                   const std::string& netmask = "255.255.255.0",
                                   const std::string& gateway = "");

int reflex_change_sensor_mode(const std::string& ip_address, int sensor_id, int8_t mode);
int reflex_change_sensor_shutter(const std::string& ip_address, int sensor_id, uint32_t shutter);
int reflex_change_sensor_gain(const std::string& ip_address, int sensor_id, float gain);

int reflex_receive_image(const std::string& ip_address, int sensor_id, ReflexImage& image);
int reflex_receive_images(const std::string& ip_address, const std::vector<int>& sensor_ids, std::vector<ReflexImage>& images);
int reflex_receive_images_full(const std::string& ip_address, std::vector<ReflexImage>& images);

int reflex_receive_data(const std::string& ip_address, ReflexData& data);
int reflex_receive_data_and_image(const std::string& ip_address, ReflexData& data, int sensor_id, ReflexImage& image);
int reflex_receive_data_and_images(const std::string& ip_address, ReflexData& data, const std::vector<int>& sensor_ids,  std::vector<ReflexImage>& images);

int reflex_receive_system_status(const std::string& ip_address, int system_id, ReflexSystemStatus& status);
int reflex_receive_sensor_status(const std::string& ip_address, int sensor_id, ReflexSensorStatus& status);

int reflex_receive_detector_parameter(const std::string& ip_address, const std::string& name, std::string& value);
int reflex_change_detector_parameter(const std::string& ip_address, const std::string& name, const std::string& value);


/*******************************************************************************
 * Helper functions/structures, not to be used publically
 *******************************************************************************/

struct ReflexRawData
{
    double                 timestamp;
    float                  framerate;
    int16_t                error;
    uint32_t               frame_id;

    std::vector<uint16_t>  contours_size;
    std::vector<int16_t>   contours_x;
    std::vector<int16_t>   contours_y;

    std::vector<float>     refl_pos2d_x;
    std::vector<float>     refl_pos2d_y;
    std::vector<float>     refl_pos3d_x;
    std::vector<float>     refl_pos3d_y;
    std::vector<float>     refl_pos3d_z;
    std::vector<float>     refl_area;
    std::vector<float>     refl_circularity;
    std::vector<float>     refl_score;

    std::vector<float>     obj_pos2d_x;
    std::vector<float>     obj_pos2d_y;
    std::vector<float>     obj_pos3d_x;
    std::vector<float>     obj_pos3d_y;
    std::vector<float>     obj_pos3d_z;
    std::vector<float>     obj_vel3d_x;
    std::vector<float>     obj_vel3d_y;
    std::vector<float>     obj_vel3d_z;
    std::vector<float>     obj_score;
    std::vector<uint32_t>  obj_lifetime;
    std::vector<int16_t>   obj_id;

    ReflexRawData()
    {
        timestamp  = -1.;
        framerate  = -1.;
        error      = -1;
        frame_id   = 0;
    }
};


int reflex_receive_data(const std::string& ip_address, ReflexRawData& data);
int reflex_extract_data(const ReflexRawData& raw_data, ReflexData& data);

int _extract_server_error(char* data, int32_t data_size);


} // end of namespace reflex



#endif
