#ifndef IMAGE_REGION_H
#define IMAGE_REGION_H

#include <reflex/base/FileAccessObject.hpp>


namespace reflex
{

template<typename N>
class Region2 : public FileAccessObject
{

public:

    enum types 
    { 
        IMAGE_REGION_UNDEFINED = -1,
        IMAGE_REGION_RECTANGLE, 
        IMAGE_REGION_POLYGONE, 
        IMAGE_REGION_CIRCLE 
    };

    Region2(int type);
    virtual ~Region2() {}

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    static cv::Ptr<Region2> create(int type);
    static cv::Ptr<Region2> create(const cv::FileNode& node);
    static cv::Ptr<Region2> create(const string& filename);

    virtual double getArea() const = 0;
    virtual cv::Size_<N> getSize() const = 0;
    virtual N getWidth() const = 0;
    virtual N getHeight() const = 0;
    virtual cv::Rect_<N> getBoundingRect() const = 0;
    virtual cv::Point_<N> getCenter() const = 0;
    virtual void move(const cv::Point_<N>& offset) = 0;

    virtual bool contains(const cv::Point_<N>& point) const = 0;
    virtual bool contains(const cv::Vec<N,2>& point) const { return contains(cv::Point_<N>(point)); }

protected:

};

typedef Region2<int> Region2i;
typedef Region2<float> Region2f;
typedef Region2<double> Region2d;


/************************************************************************/


template<typename N>
class RegionRect2 : public Region2<N>
{
public:

    RegionRect2();
    RegionRect2(N origX, N origY, N width, N height);
    RegionRect2(const cv::Rect_<N>& rect);
    RegionRect2(const cv::Point_<N>& origin, const cv::Size_<N>& size);
    virtual ~RegionRect2() {}

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual double getArea() const { return rect.width * rect.height; }
    virtual N getWidth() const { return rect.width; }
    virtual N getHeight() const { return rect.height; }
    virtual cv::Size_<N> getSize() const { return cv::Size_<N>(rect.width, rect.height); }
    virtual cv::Rect_<N> getBoundingRect() const { return rect; }
    virtual cv::Point_<N> getCenter() const { return cv::Point_<N>(rect.x + 0.5 * rect.width, rect.y + 0.5 * rect.height); }
    virtual cv::Point_<N> getOrigin() const { return cv::Point_<N>(rect.x, rect.y); }

    virtual void move(const cv::Point_<N>& offset) { rect.x += offset.x; rect.y += offset.y; }

    virtual bool contains(const cv::Point_<N>& point) const { return point.inside(rect); }

    cv::Rect_<N> rect;

protected:

    template <class T>
    static inline bool loadField(const cv::FileNode& node, T& value)
    {
        if(node.empty())
            return false;

        node >> value;
        return true;
    }

    template <class T>
    static inline void saveField(cv::FileStorage& fs, string key, T& value)
    {
        fs << key << value;
    }
};

typedef RegionRect2<int> RegionRect2i;
typedef RegionRect2<float> RegionRect2f;
typedef RegionRect2<double> RegionRect2d;


/************************************************************************/


template<typename N>
class RegionPolygone2 : public Region2<N>
{
public:

    RegionPolygone2();
    RegionPolygone2(const std::vector<cv::Point_<N>>& points);
    virtual ~RegionPolygone2() {}

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual double getArea() const;
    virtual N getWidth() const { return getBoundingRect().width; }
    virtual N getHeight() const { return getBoundingRect().height; }
    virtual cv::Size_<N> getSize() const;
    virtual cv::Rect_<N> getBoundingRect() const;
    virtual cv::Point_<N> getCenter() const;
    virtual void move(const cv::Point_<N>& offset);

    virtual bool contains(const cv::Point_<N>& point) const;

    std::vector<cv::Point_<N>> points;

protected:

    template <class T>
    static inline bool loadField(const cv::FileNode& node, T& value)
    {
        if(node.empty())
            return false;

        node >> value;
        return true;
    }

    template <class T>
    static inline void saveField(cv::FileStorage& fs, string key, T& value)
    {
        fs << key << value;
    }
};


typedef RegionPolygone2<int> RegionPolygone2i;
typedef RegionPolygone2<float> RegionPolygone2f;
typedef RegionPolygone2<double> RegionPolygone2d;


/************************************************************************/


template<typename N>
class RegionCircle2 : public Region2<N>
{
public:

    RegionCircle2();
    RegionCircle2(N centerX, N centerY, N radius);
    RegionCircle2(const cv::Point_<N>& center, N radius);
    virtual ~RegionCircle2() {}

    virtual void write(cv::FileStorage& fs) const;
    virtual void read(const cv::FileNode& node);

    virtual double getArea() const { return radius * radius * M_PI; }
    virtual N getWidth() const { return 2 * radius; }
    virtual N getHeight() const { return 2 * radius; }
    virtual cv::Size_<N> getSize() const { return cv::Size_<N>(2*radius,2*radius); }
    virtual cv::Rect_<N> getBoundingRect() const { return cv::Rect_<N>(center.x-radius,center.y-radius,2*radius, 2*radius); }
    virtual cv::Point_<N> getCenter() const { return center; }
    virtual void move(const cv::Point_<N>& offset) { center += offset; }

    virtual bool contains(const cv::Point_<N>& point) const;

    cv::Point_<N> center;
    N radius;

protected:

    template <class T>
    static inline bool loadField(const cv::FileNode& node, T& value)
    {
        if(node.empty())
            return false;

        node >> value;
        return true;
    }

    template <class T>
    static inline void saveField(cv::FileStorage& fs, string key, T& value)
    {
        fs << key << value;
    }
};


typedef RegionCircle2<int> RegionCircle2i;
typedef RegionCircle2<float> RegionCircle2f;
typedef RegionCircle2<double> RegionCircle2d;


} // end namespace reflex



namespace cv
{
    void write(cv::FileStorage& fs, const string&, const reflex::RegionRect2f&);
}

#endif
