#ifndef REFLEX_H
#define REFLEX_H

#include <reflex/driver/reflex.hpp>
#include <opencv2/opencv.hpp>

#define REFLEX_VERSION_MAJOR 2
#define REFLEX_VERSION_MINOR 1
#define REFLEX_VERSION_PATCH 6

using namespace std;

namespace reflex
{


enum
{
    CAM_ALL = -1,
    CAM1 = 0,
    CAM2,
    CAM3,
    CAM4
};


enum
{
    SYSTEM_MASTER = 0,
    SYSTEM_SLAVE = 1
};


inline std::string getEnvironmentVariable(const char* name)
{
    std::string val;
    if(const char* env_p = std::getenv(name))
        val.assign(env_p);

    return val;
}


inline std::string getEnvironmentVariable(const std::string& name)
{
    return getEnvironmentVariable(name.c_str());
}


inline std::string errorMessage(int errorCode)
{
    switch(errorCode)
    {
    case ERROR_NONE:
        return std::string("None");
    case ERROR_HARDWARE:
        return std::string("Hardware Error");
    case ERROR_NETWORK:
        return std::string("Network Error");
    case ERROR_TIME_OUT:
        return std::string("Timeout Error");
    case ERROR_MEMORY:
        return std::string("Out of memory");
    case ERROR_BOUNDS:
        return std::string("Out of bounds");
    case ERROR_DIMENSIONS:
        return std::string("Dimensionality Error");
    case ERROR_INITIALIZATION:
        return std::string("Initialization Error");
    case ERROR_DATA_EMPTY:
        return std::string("Empty Data Error");
    case ERROR_DATA_TYPE:
        return std::string("Data Type Error");
    case ERROR_IMAGE_TYPE:
        return std::string("Image Type Error");
    case ERROR_IMAGE_SIZE:
        return std::string("Image Size Error");
    case ERROR_IMAGE_CONTENT:
        return std::string("Image Content Error");
    case ERROR_FILE_ACCESS_READ:
        return std::string("File Read Access Error");
    case ERROR_FILE_ACCESS_WRITE:
        return std::string("File Write Access Error");
    case ERROR_FILE_ACCESS_EXECUTE:
        return std::string("File Execution Error");
    case ERROR_XML_EMPTY_ROOT_NODE:
        return std::string("Empty XML Root Node");
    case ERROR_XML_EMPTY_NODE:
        return std::string("Empty XML Node");
    case ERROR_INVALID_COMMAND:
        return std::string("Invalid Command");
    case ERROR_UNKNOWN_COMMAND:
        return std::string("Unknown Command");
    case ERROR_INVALID_PARAM:
        return std::string("Invalid Parameter");
    case ERROR_CAMERA_TRIGGER:
        return std::string("Camera Trigger Error");
    case ERROR_CALIBRATION:
        return std::string("Calibration Error");
    case ERROR_SERVER:
        return std::string("Server Error");
    case ERROR_DETECTOR:
        return std::string("Detector Error");
    case ERROR_CONFIGURATION:
        return std::string("Configuration Error");
    case ERROR_BEEPER_ACCESS:
        return std::string("Beeper Access Error");
    case ERROR_IOCTL:
        return std::string("IOCTL Error");
    case ERROR_SOUND_SUPPORT:
        return std::string("Sound Support Error");
    case ERROR_SOUND_PARAMETER:
        return std::string("Sound Parameter not found");
    case ERROR_NOT_ALLOWED:
        return std::string("Not allowed");
    case ERROR_NOT_IMPLEMENTED:
        return std::string("Not implemented");
    case ERROR_BROKEN_SEND:
        return std::string("Data Transmission Error");
    default:
        std::stringstream x;
        x << "Error " << errorCode;
        return x.str();
    }
}


inline string getErrorMessage(int errorCode, const string& msg = string())
{
    stringstream res;
    res << "Error: " << errorMessage(errorCode) << " (ERR " << errorCode << ")";
    if(!msg.empty())
        res << ": " << msg;

    return res.str();
}


inline string getErrorMessage(const string& objectTag, int errorCode, const string& msg = string())
{
    stringstream res;
    res << "Error in " << objectTag << ": " << errorMessage(errorCode) << " (ERR " << errorCode << ")";
    if(!msg.empty())
        res << ": " << msg;

    return res.str();
}


inline string getErrorMessage(const string& objectTag, const string& function, int errorCode, const string& msg = string())
{
    stringstream res;
    res << "Error in " << objectTag << "::" << function << "(): " << errorMessage(errorCode) << " (ERR " << errorCode << ")";
    if(!msg.empty())
        res << ": " << msg;

    return res.str();
}


inline void printErrorMessage(int errorCode, const string& msg = string())
{
    cout << "\033[31m" << getErrorMessage(errorCode, msg) << "\033[0m\n";
}


inline void printErrorMessage(const string& objectTag, int errorCode, const string& msg = string())
{
    cout << "\033[31m" << getErrorMessage(objectTag, errorCode, msg) << "\033[0m\n";
}


inline void printErrorMessage(const string& objectTag, const string& function, int errorCode, const string& msg = string())
{
    cout << "\033[31m" << getErrorMessage(objectTag, function, errorCode, msg) << "\033[0m\n";
}


inline string getWarningMessage(const string& objectTag, int errorCode, const string& msg = string())
{
    stringstream res;
    res << objectTag << ": Warning " << errorCode << " (" << errorMessage(errorCode) << ")";
    if(!msg.empty())
        res << ": " << msg;

    return res.str();
}


inline string getWarningMessage(const string& objectTag, const string& function, int errorCode, const string& msg = string())
{
    stringstream res;
    res << "Warning " << errorCode << " in " << objectTag << "::" << function << "(): " << errorMessage(errorCode);
    if(!msg.empty())
        res << ": " << msg;

    return res.str();
}

inline void printWarningMessage(const string& objectTag, int errorCode, const string& msg = string())
{
    cout << "\033[33m" << getWarningMessage(objectTag, errorCode, msg) << "\033[0m\n";
}


inline void printWarningMessage(const string& objectTag, const string& function, int errorCode, const string& msg = string())
{
    cout << "\033[33m" << getWarningMessage(objectTag, function, errorCode, msg) << "\033[0m\n";
}


} // end of namespace reflex


namespace cv
{

void info(const std::string& name, const cv::Mat& mat);

void print(const cv::Mat mat, int digits, int precision);


template <class _Tp>
inline void write(cv::FileStorage& fs, const cv::String& key, std::vector<cv::Point_<_Tp>> value)
{
    fs << "[:";
    for(typename vector<cv::Point_<_Tp>>::const_iterator it = value.begin(); it != value.end(); ++it)
        fs << *it;
    fs << "]";
}


template <class _Tp>
inline void write(cv::FileStorage& fs, const cv::String& key, std::vector<std::vector<cv::Point_<_Tp>>> value)
{
    fs << "[:";
    for(typename vector<vector<cv::Point_<_Tp>>>::const_iterator it = value.begin(); it != value.end(); ++it)
        fs << *it;
    fs << "]";
}


template <class _Tp>
inline void write(cv::FileStorage& fs, const cv::String& key, std::vector<cv::Point3_<_Tp>> value)
{
    fs << "[:";
    for(typename vector<cv::Point3_<_Tp>>::const_iterator it = value.begin(); it != value.end(); ++it)
        fs << *it;
    fs << "]";
}


template <class _Tp>
inline void write(cv::FileStorage& fs, const cv::String& key, std::vector<std::vector<cv::Point3_<_Tp>>> value)
{
    fs << "[:";
    for(typename vector<vector<cv::Point3_<_Tp>>>::const_iterator it = value.begin(); it != value.end(); ++it)
        fs << *it;
    fs << "]";
}


template <class _Tp1, int cn>
inline void write(cv::FileStorage& fs, const cv::String& key, std::vector<cv::Vec<_Tp1,cn>> value)
{
    fs << "[:";
    for(typename vector<cv::Vec<_Tp1,cn>>::const_iterator it = value.begin(); it != value.end(); ++it)
        fs << *it;
    fs << "]";
}


} // end of namespace cv

#endif
