#ifndef PARTICLE_FILTER_H
#define PARTICLE_FILTER_H

#include <reflex/reflex.hpp>

namespace reflex
{

class ParticleFilter
{
public:

    class Particle
    {
    public:

        Particle();
        double weight;
        cv::Mat state;
    };


    class Resampler
    {
    public:
        Resampler() {};
        virtual ~Resampler() {};
        virtual void process(vector<Particle>& particles) = 0;
        virtual void process(vector<Particle>& particles, const cv::Mat& measurement) = 0;
    };


    class SystematicResampler : public Resampler
    {
    public:

        virtual void process(vector<Particle>& particles);
        virtual void process(vector<Particle>& particles, const cv::Mat& measurement);
    };


    class MotionModel
    {
    public:
        virtual void process(vector<Particle>& particles, double timeStep) = 0;
    };


    class MeasurementModel
    {
    public:
        virtual void process(vector<Particle> &particles, const cv::Mat &measurement, double& minWeight, double& maxWeight) = 0;
    };


    class Parameters
    {
    public:
        Parameters();

        int dimState;
        int nbParticles;
        cv::Mat stateMin;
        cv::Mat stateMax;
    };

    ParticleFilter();
    ParticleFilter(const Parameters& parameters);

    virtual ~ParticleFilter();

    virtual int init(const Parameters& parameters);

    int resample();
    int predict(double timeStep);
    int correct(const cv::Mat& measurement);

    cv::Mat getStateMaximum(int* idx = NULL, double* weight = NULL) const;
    cv::Mat getStateWeightedMean(float ratio = 1.) const;
    cv::Mat getStateRobustMean(double alpha, int* nbParticles = NULL) const;

    void normalizeWeights();

    void initializeParticleUniform(Particle& particle);
    void initializeParticleNormal(Particle& particle, const cv::Mat& stateMean, const cv::Mat& stateVariance);

    static double getRandomNormal(double mean = 0.0, double sigma = 1.0);
    static int getRandomUniform(int valMin, int valMax);


    cv::Ptr<Resampler> resampler;
    cv::Ptr<MotionModel> motionModel;
    cv::Ptr<MeasurementModel> measurementModel;

    vector<Particle> particles;
    Parameters params;

    double weightMax;
    double weightMin;

    int idxMax;
    int idxMin;

protected:

    static cv::RNG _rng;

    // helper vectors
    cv::Mat _v1;
    cv::Mat _v2;

};

}

#endif
