#include "ros/ros.h"
#include "tf/tf.h"
#include "std_msgs/String.h"
#include "reflex_ros/SetSensorMode.h"
#include "reflex_ros/SetExposure.h"
#include <reflex/driver/ReflexDriverDev.hpp>
#include <reflex/application/ReflexCamera.hpp>
#include <reflex/camera/SensorConfig.hpp>
#include <reflex/camera/CameraIntrinsics.hpp>
#include <reflex/camera/CameraExtrinsics.hpp>
#include <reflex/camera/ImageRectifier.hpp>
#include <reflex/system/UnitInfo.hpp>
#include <sstream>
#include <cv_bridge/cv_bridge.h>
#include <image_transport/image_transport.h>
#include <camera_info_manager/camera_info_manager.h>
#include <geometry_msgs/PoseStamped.h>

#include "reflex_ros/ReflexCameraUnit.h"
#include "reflex_ros/ReflexData.h"
#include "reflex_ros/ReflexImage.h"
#include "reflex_ros/ReflexObject.h"
#include "reflex_ros/ReflexReflector.h"
#include "reflex_ros/ReflexSensorStatus.h"
#include "reflex_ros/ReflexSystemStatus.h"
#include "reflex_ros/ReflexContour.h"



using namespace reflex;

ReflexCamera                _cameraUnit;
int                         _cnt = 0;
ros::Time                   _startTimeROS;
double                      _startTimeRefleX;
int                         _defaultMode = SYSTEM_STATE_IDLE;
bool                        _drawDetections = false;
const string                _dirCalib = "/tmp/";
int                         _operationMode = SYSTEM_STATE_UNDEFINED;
string                      _unitSerialNumber;
string                      _cameraIPAddress = "192.168.3.15";
sensor_msgs::CameraInfoPtr  _msgCameraInfo[2];
vector<ReflexSystemStatus>  _systemStatus;
int                         _disconnect = 0;
float                       _timeout = 5.;


int setCameraMode(int mode)
{
    string txtMode;
    switch(mode)
    {
    case SYSTEM_STATE_IDLE:
        txtMode = "IDLE";
        break;
    case SYSTEM_STATE_DETECTION:
        txtMode = "DETECTION";
        break;
    case SYSTEM_STATE_RAW_DATA:
        txtMode = "RAW_DATA";
        break;
    }

    int err = _cameraUnit.changeOperationState(mode, _timeout);
    if(err != ERROR_NONE)
    {
        ROS_ERROR("Failed to Change Operation Mode to %s (Error %d)", txtMode.c_str(), err);
        return ERROR_SERVER;
    }

    _operationMode = mode;
    _cnt = 0;

    ROS_INFO("Change Operation Mode to %s", txtMode.c_str());

    return ERROR_NONE;
}


int checkCameraState()
{
    ReflexCameraUnit cameraUnitInfo;
    int err = reflex_receive_camera_unit_info(_cameraIPAddress, cameraUnitInfo);
    if(err != ERROR_NONE)
    {
        //log(QString::fromStdString("Failed to read unit serial number"));

        if(!_unitSerialNumber.empty())
        {
            if(_disconnect < 5)
                _disconnect++;
            else
            {
                ROS_INFO("Camera Unit %s disconnected", _unitSerialNumber.c_str());
                _unitSerialNumber.clear();

                _operationMode = SYSTEM_STATE_UNDEFINED;

                _disconnect = 0;
            }
        }
        return err;
    }


    if(cameraUnitInfo.serial_number != _unitSerialNumber)
    {
        _unitSerialNumber = cameraUnitInfo.serial_number;

        ROS_INFO("Camera Unit %s connected", _unitSerialNumber.c_str());

        _cameraUnit = ReflexCamera(_cameraIPAddress, _unitSerialNumber);
        err = _cameraUnit.init(_dirCalib);
        if(err != ERROR_NONE)
        {
            ROS_ERROR("Failed to initialize camera unit (Error %i)", err);
            return err;
        }
    }

    // activation
    if(_operationMode == SYSTEM_STATE_UNDEFINED)
        setCameraMode(_defaultMode);

    err = _cameraUnit.receiveSystemStatus(_systemStatus);
    if(err != ERROR_NONE)
    {
        ROS_ERROR("Failed to read system status (Error %i)", err);
        return err;
    }


    return ERROR_NONE;
}


void convertToROS(const ReflexSystemStatus& status, reflex_ros::ReflexSystemStatus& msgStatus)
{
    msgStatus.header.stamp = ros::Time(status.timestamp);
    msgStatus.temperature = status.temperature;
    msgStatus.error = status.error;
    msgStatus.state = status.state;
}

void convertToROS(const ReflexSensorStatus& status, reflex_ros::ReflexSensorStatus& msgStatus)
{
    msgStatus.header.stamp = ros::Time(status.timestamp);
    msgStatus.sensor_id = status.sensor_id;
    msgStatus.error = status.error;
    msgStatus.frame_id = status.frame_id;
    msgStatus.framerate = status.framerate;
    msgStatus.exposure_mode = status.exposure_mode;
    msgStatus.shutter = status.shutter;
    msgStatus.gain = status.gain;
}


void convertToROS(const ReflexData& data, reflex_ros::ReflexData& msgData)
{
    msgData.reflectors.clear();
    msgData.objects.clear();

    msgData.frame_id = data.frame_id;
    msgData.error = data.error;
    msgData.framerate = data.framerate;

    for(int k = 0; k < data.reflectors.size(); k++)
    {
        reflex_ros::ReflexReflector reflector;
        reflector.pos2D.x     = data.reflectors[k].pos2d[X];
        reflector.pos2D.y     = data.reflectors[k].pos2d[Y];
        reflector.pos3D.x     = data.reflectors[k].pos3d[X];
        reflector.pos3D.y     = data.reflectors[k].pos3d[Y];
        reflector.pos3D.z     = data.reflectors[k].pos3d[Z];
        reflector.score       = data.reflectors[k].score;
        reflector.circularity = data.reflectors[k].circularity;
        reflector.area        = data.reflectors[k].area;
        reflector.contour.x   = data.reflectors[k].contour.x;
        reflector.contour.y   = data.reflectors[k].contour.y;

        msgData.reflectors.push_back(reflector);
    }

    for(int k = 0; k < data.objects.size(); k++)
    {
        reflex_ros::ReflexObject object;
        object.pos2D.x    = data.objects[k].pos2d[X];
        object.pos2D.y    = data.objects[k].pos2d[Y];
        object.pos3D.x    = data.objects[k].pos3d[X];
        object.pos3D.y    = data.objects[k].pos3d[Y];
        object.pos3D.z    = data.objects[k].pos3d[Z];
        object.vel3D.x    = data.objects[k].vel3d[X];
        object.vel3D.y    = data.objects[k].vel3d[Y];
        object.vel3D.z    = data.objects[k].vel3d[Z];
        object.score      = data.objects[k].score;
        object.id         = data.objects[k].id;
        object.lifetime   = data.objects[k].lifetime;

        msgData.objects.push_back(object);
    }
}


bool setSensorMode(reflex_ros::SetSensorMode::Request &req, reflex_ros::SetSensorMode::Response &res)
{
    int err = setCameraMode(req.mode);
    res.err = err;
    if(err != ERROR_NONE)
        return false;

    return true;
}


int setExposure(bool mode, int shutter, float gain)
{
    int err = ERROR_NONE;
    int val = (mode) ? SensorConfig::EXPOSURE_TIME_AUTO : SensorConfig::EXPOSURE_TIME_FIXED;

    if(mode == true)
    {
        ROS_WARN("NIR Cameras do not support auto exposure.");
        mode = false;
        val = SensorConfig::EXPOSURE_TIME_FIXED;
    }
    else
    {
        int err = reflex_change_sensor_mode(_cameraIPAddress, SENSOR_ALL, val);
        if(err != ERROR_NONE)
        {
            ROS_WARN("Failed to change exposure mode: Error %d", err);
            return err;
        }
    }

    if(val == SensorConfig::EXPOSURE_TIME_FIXED)
    {
        err = reflex_change_sensor_shutter(_cameraIPAddress, SENSOR_ALL, shutter);
        if(err != ERROR_NONE)
        {
            ROS_WARN("Failed to change exposure time: Error %d", err);
            return err;
        }

        err = reflex_change_sensor_gain(_cameraIPAddress, SENSOR_ALL, gain);
        if(err != ERROR_NONE)
        {
            ROS_WARN("Failed to change camera gain: Error %d", err);
            return err;
        }
    }

    return ERROR_NONE;
}


bool setExposure(reflex_ros::SetExposure::Request &req, reflex_ros::SetExposure::Response &res)
{
    int err = setExposure(req.automode, req.shutter, req.gain);
    res.err = err;
    if(err != ERROR_NONE)
        return false;

    return true;
}


void drawImage(cv::Mat& img, const ReflexRawData& data)
{
    int contourIdx = 0;
    for(size_t i = 0; i < data.contours_size.size(); i++)
    {
        cv::Point start_point(data.contours_x[contourIdx], data.contours_x[contourIdx]);
        cv::Point pt1 = start_point;
        for(size_t j = 1; j < data.contours_size[i]; j++)
        {
          cv::Point pt2(data.contours_x[contourIdx+j], data.contours_y[contourIdx+j]);
          cv::line(img, pt1, pt2, cv::Scalar(0,0,200), 2);
          pt1 = pt2;
        }
        cv::line(img, pt1, start_point, cv::Scalar(0,0,200), 2);
        contourIdx += data.contours_size[i];
    }
}


void getPoseMessage(const cv::Mat& R, const cv::Mat& T, geometry_msgs::PoseStamped& pose)
{
    tf::Matrix3x3 tfRot(
        R.at<double>(0,0), R.at<double>(0,1), R.at<double>(0,2),
        R.at<double>(1,0), R.at<double>(1,1), R.at<double>(1,2),
        R.at<double>(2,0), R.at<double>(2,1), R.at<double>(2,2)
    );

    tf::Quaternion tfQuat;
    tfRot.getRotation(tfQuat);

    pose.pose.orientation.x = tfQuat.getX();
    pose.pose.orientation.y = tfQuat.getY();
    pose.pose.orientation.z = tfQuat.getZ();
    pose.pose.orientation.w = tfQuat.getW();

    pose.pose.position.x = T.at<double>(X);
    pose.pose.position.y = T.at<double>(Y);
    pose.pose.position.z = T.at<double>(Z);
}


int main(int argc, char **argv)
{
    ros::init(argc, argv, "reflex_driver");
    ros::NodeHandle nh("~");

    // Read parameters

    nh.param("default_mode", _defaultMode, 0);
    nh.param("draw_detections", _drawDetections, false);

    // Advertise Topics

    _msgCameraInfo[SENSOR1] = sensor_msgs::CameraInfoPtr(new sensor_msgs::CameraInfo());
    _msgCameraInfo[SENSOR2] = sensor_msgs::CameraInfoPtr(new sensor_msgs::CameraInfo());

    _msgCameraInfo[SENSOR1]->header.frame_id = "reflex_cam1";
    _msgCameraInfo[SENSOR2]->header.frame_id = "reflex_cam2";

    // Advertise NIR Images with Flash (signal images)

    image_transport::ImageTransport itNIRsig1(nh);
    image_transport::ImageTransport itNIRsig2(nh);
    image_transport::CameraPublisher pubCamNIRsig[2];
    pubCamNIRsig[CAM1] = itNIRsig1.advertiseCamera("cam1/sig/image", 1);
    pubCamNIRsig[CAM2] = itNIRsig2.advertiseCamera("cam2/sig/image", 1);

    // Advertise NIR Images without Flash (reference images)

    image_transport::ImageTransport itNIRref1(nh), itNIRref2(nh);
    image_transport::CameraPublisher pubCamNIRref[2];
    pubCamNIRref[CAM1] = itNIRref1.advertiseCamera("cam1/ref/image", 1);
    pubCamNIRref[CAM2] = itNIRref2.advertiseCamera("cam2/ref/image", 1);

    // Advertise Detections

    reflex_ros::ReflexData msgData;
    msgData.header.frame_id = "reflex_cam1";

    ros::Publisher pubData = nh.advertise<reflex_ros::ReflexData>("cam1/detections", 1);

    // Advertise System State

    reflex_ros::ReflexSystemStatus msgSystemStatus;
    vector<ros::Publisher> pubSystemStatus;

    // Advertise transformations between cameras

    ros::Publisher pubPoseNIR;
    pubPoseNIR = nh.advertise<geometry_msgs::PoseStamped>("extrinsics/cam1_cam2", 100);

    // Advertise Services

    ros::ServiceServer srvSensorMode = nh.advertiseService("set_operation_state", setSensorMode);
    ros::ServiceServer srvExposure = nh.advertiseService("set_exposure", setExposure);

    int err = ERROR_NONE;

    while (ros::ok())
    {
        ros::Time tLoop = ros::Time::now();

        cv::Mat imgNIRSig[2];
        cv::Mat imgNIRRef[2];

        ReflexData data;

        if(_operationMode == SYSTEM_STATE_UNDEFINED)
        {
            err = checkCameraState();
            if(err != ERROR_NONE)
            {
                ROS_WARN("No RefleX Camera Unit connected.");
                ros::Duration(2.).sleep();
            }

            if(_operationMode != SYSTEM_STATE_UNDEFINED)
            {
                pubSystemStatus.resize(_cameraUnit.unitInfo.systems.size());
                for(size_t sysID = 0; sysID < _cameraUnit.unitInfo.systems.size(); sysID++)
                {
                    stringstream statusTopicName;
                    statusTopicName << "sys" << sysID+1 << "/system_status";
                    pubSystemStatus[sysID] = nh.advertise<reflex_ros::ReflexSystemStatus>(statusTopicName.str(), 1);
                }
            }
        }
        else if(_operationMode == SYSTEM_STATE_IDLE)
        {
            checkCameraState();

            for(size_t sysID = 0; sysID < _cameraUnit.unitInfo.systems.size(); sysID++)
            {
                convertToROS(_systemStatus[sysID], msgSystemStatus);
                pubSystemStatus[sysID].publish(msgSystemStatus);
            }
        }
        else if(_operationMode == SYSTEM_STATE_DETECTION)
        {
            std::vector<ReflexImageCV> images;
            std::vector<int> sensorIDs = { SENSOR1, SENSOR2 };
            err = reflex_receive_data_and_images(_cameraIPAddress, data, sensorIDs, images);
            if(err != ERROR_NONE && err != ERROR_DATA_EMPTY)
            {
                ROS_ERROR("Failed to receive data and images (Error %i)", err);
                continue;
            }

            if(err == ERROR_NONE)
            {
                if(_cnt == 0)
                {
                    _startTimeROS = ros::Time::now();
                    _startTimeRefleX = data.timestamp;
                }

                if(_cnt % 10 == 0)
                {
                    checkCameraState();

                    for(size_t sysID = 0; sysID < _cameraUnit.unitInfo.systems.size(); sysID++)
                    {
                        convertToROS(_systemStatus[sysID], msgSystemStatus);
                        pubSystemStatus[sysID].publish(msgSystemStatus);
                    }
                }

                ros::Time timeStamp = _startTimeROS + ros::Duration(data.timestamp-_startTimeRefleX);

                sensor_msgs::ImagePtr msgNIRsig1 = cv_bridge::CvImage(std_msgs::Header(), "mono8", images[CAM1].img).toImageMsg();
                msgNIRsig1->header.stamp = timeStamp;
                msgNIRsig1->header.frame_id = "reflex_cam1";
                _msgCameraInfo[CAM1]->header.stamp = timeStamp;
                pubCamNIRsig[CAM1].publish(msgNIRsig1, _msgCameraInfo[CAM1]);

                sensor_msgs::ImagePtr msgNIRsig2 = cv_bridge::CvImage(std_msgs::Header(), "mono8", images[CAM2].img).toImageMsg();
                msgNIRsig2->header.stamp = timeStamp;
                msgNIRsig2->header.frame_id = "reflex_cam2";
                _msgCameraInfo[CAM2]->header.stamp = timeStamp;
                pubCamNIRsig[CAM2].publish(msgNIRsig2, _msgCameraInfo[CAM2]);

                if(pubData.getNumSubscribers() > 0)
                {
                    // Convert detections to ROS format and publish
                    convertToROS(data, msgData);
                    msgData.header.stamp = timeStamp;
                    pubData.publish(msgData);
                }

                geometry_msgs::PoseStamped msgPoseNIR;
                getPoseMessage(_cameraUnit.extrinsics[CAM1][CAM2].R, _cameraUnit.extrinsics[CAM1][CAM2].T, msgPoseNIR);
                msgPoseNIR.header.stamp = timeStamp;
                pubPoseNIR.publish(msgPoseNIR);


                _cnt++;
            }

        }
        else if(_operationMode == SYSTEM_STATE_RAW_DATA)
        {
            std::vector<ReflexImageCV> images;
            err = reflex_receive_images_full(_cameraIPAddress, images);
            if(err == ERROR_NONE)
            {
                if(_cnt == 0)
                {
                    _startTimeROS = ros::Time::now();
                    _startTimeRefleX = images[0].timestamp;
                }

                if(_cnt % 10 == 0)
                {
                    checkCameraState();

                    for(size_t sysID = 0; sysID < _cameraUnit.unitInfo.systems.size(); sysID++)
                    {
                        convertToROS(_systemStatus[sysID], msgSystemStatus);
                        pubSystemStatus[sysID].publish(msgSystemStatus);
                    }
                }

                ros::Time tsNIRsig = _startTimeROS + ros::Duration(images[0].timestamp-_startTimeRefleX);
                ros::Time tsNIRref = _startTimeROS + ros::Duration(images[2].timestamp-_startTimeRefleX);

                sensor_msgs::ImagePtr msgNIRsig1 = cv_bridge::CvImage(std_msgs::Header(), "mono8", imgNIRSig[CAM1]).toImageMsg();
                msgNIRsig1->header.stamp = tsNIRsig;
                msgNIRsig1->header.frame_id = "reflex_nir_cam1";
                _msgCameraInfo[CAM1]->header.stamp = tsNIRsig;
                pubCamNIRsig[CAM1].publish(msgNIRsig1, _msgCameraInfo[CAM1]);

                sensor_msgs::ImagePtr msgNIRsig2 = cv_bridge::CvImage(std_msgs::Header(), "mono8", imgNIRSig[CAM2]).toImageMsg();
                msgNIRsig2->header.stamp = tsNIRsig;
                msgNIRsig2->header.frame_id = "reflex_nir_cam2";
                _msgCameraInfo[CAM2]->header.stamp = tsNIRsig;
                pubCamNIRsig[CAM2].publish(msgNIRsig2, _msgCameraInfo[CAM2]);

                sensor_msgs::ImagePtr msgNIRref1 = cv_bridge::CvImage(std_msgs::Header(), "mono8", imgNIRRef[CAM1]).toImageMsg();
                msgNIRref1->header.stamp = tsNIRref;
                msgNIRref1->header.frame_id = "reflex_nir_cam1";
                _msgCameraInfo[CAM1]->header.stamp = tsNIRref;
                pubCamNIRref[CAM1].publish(msgNIRref1, _msgCameraInfo[CAM1]);

                sensor_msgs::ImagePtr msgNIRref2 = cv_bridge::CvImage(std_msgs::Header(), "mono8", imgNIRRef[CAM2]).toImageMsg();
                msgNIRref2->header.stamp = tsNIRref;
                msgNIRref2->header.frame_id = "reflex_nir_cam2";
                _msgCameraInfo[CAM2]->header.stamp = tsNIRref;
                pubCamNIRref[CAM2].publish(msgNIRref2, _msgCameraInfo[CAM2]);

                geometry_msgs::PoseStamped msgPoseNIR;
                getPoseMessage(_cameraUnit.extrinsics[CAM1][CAM2].R, _cameraUnit.extrinsics[CAM1][CAM2].T, msgPoseNIR);
                msgPoseNIR.header.stamp = tsNIRsig;
                pubPoseNIR.publish(msgPoseNIR);

                _cnt++;
            }
        }

        ros::spinOnce();

        ros::Duration dT = ros::Time::now() - tLoop;
        if(dT.toSec() < 0.05)
            ros::Duration(0.05 - dT.toSec()).sleep();

    }

    setCameraMode(SYSTEM_STATE_IDLE);

    return 0;
}
